library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.OptoUtils.all;

entity Bram2x is
	generic (
		width : natural := 64;
		adr_width : natural := 10
	 );
    Port (
         clk : IN STD_LOGIC;
				 reset : IN STD_LOGIC;
				 we : IN STD_LOGIC_VECTOR(1 downto 0);
				 adr_r : IN STD_LOGIC_VECTOR(2*adr_width-1 DOWNTO 0);
				 adr_w : IN STD_LOGIC_VECTOR(2*adr_width-1 DOWNTO 0);
				 din : IN STD_LOGIC_VECTOR(2*width-1 DOWNTO 0);
				 dout : OUT STD_LOGIC_VECTOR(2*width-1 DOWNTO 0)
			  );
end Bram2x;

architecture Bram2x_body of Bram2x is
	component Bram
	generic (
	  width     : natural := 64;
	  adr_width : natural := 10
	);
	port (
	  clk   : IN  STD_LOGIC;
	  reset : IN  STD_LOGIC;
	  we    : IN  STD_LOGIC;
	  adr_r : IN  STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
	  adr_w : IN  STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
	  din   : IN  STD_LOGIC_VECTOR(width-1 DOWNTO 0);
	  dout  : OUT STD_LOGIC_VECTOR(width-1 DOWNTO 0)
	);
	end component Bram;

  signal pump_we : std_logic;
  signal pump_adr_r : STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
  signal pump_adr_w : STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
  signal pump_din  : STD_LOGIC_VECTOR(width-1 DOWNTO 0);
  signal pump_dout : STD_LOGIC_VECTOR(width-1 DOWNTO 0);

  signal port1Clk : std_logic;
  signal doutReg : STD_LOGIC_VECTOR(width-1 DOWNTO 0);

	signal adr1_r : std_logic_vector(adr_width-1 downto 0);
	signal adr2_r : std_logic_vector(adr_width-1 downto 0);
	signal adr1_w : std_logic_vector(adr_width-1 downto 0);
	signal adr2_w : std_logic_vector(adr_width-1 downto 0);
	signal data1_in : std_logic_vector(width-1 downto 0);
	signal data2_in : std_logic_vector(width-1 downto 0);

	signal we_inner : std_logic_vector(1 downto 0);
	signal swapAddrRead : std_logic;
	signal swapAddrWrite : std_logic;
	signal swapOutputRead : std_logic;
	signal mirrorOutput : std_logic;

	signal swapReadUndelayed : std_logic;
	signal swapWriteUndelayed : std_logic;
	signal mirrorOutputUndelayed : std_logic;

	signal swapReadDelayed  : std_logic_vector(1 downto 0);
  signal swapWriteDelayed : std_logic_vector(1 downto 0);
  signal mirrorOutputDelayed : std_logic_vector(1 downto 0);
begin

	Adr_railway : process(adr_r, adr_w, din, swapAddrRead, swapAddrWrite)
		variable v_adr1_r : std_logic_vector(adr_width-1 downto 0);
		variable v_adr2_r : std_logic_vector(adr_width-1 downto 0);
		variable v_adr1_w : std_logic_vector(adr_width-1 downto 0);
		variable v_adr2_w : std_logic_vector(adr_width-1 downto 0);
		variable v_data1_in : std_logic_vector(width-1 downto 0);
		variable v_data2_in : std_logic_vector(width-1 downto 0);
	begin
		v_adr1_r := adr_r(adr_width-1 downto 0);
		v_adr2_r := adr_r(2*adr_width-1 downto adr_width);
		v_adr1_w := adr_w(adr_width-1 downto 0);
		v_adr2_w := adr_w(2*adr_width-1 downto adr_width);
		v_data1_in := din(width-1 downto 0);
		v_data2_in := din(2*width-1 downto width);

		adr1_r <= v_adr1_r;
		adr2_r <= v_adr2_r;
		if(swapAddrRead = '1') then
			adr1_r <= v_adr2_r;
			adr2_r <= v_adr1_r;
		end if;

		adr1_w <= v_adr1_w;
		adr2_w <= v_adr2_w;
		data1_in <= v_data1_in;
		data2_in <= v_data2_in;
		we_inner <= we;
		if(swapAddrWrite = '1') then
			adr1_w <= v_adr2_w;
			adr2_w <= v_adr1_w;
			we_inner(1) <= we(0);
			we_inner(0) <= we(1);
			data2_in <= v_data1_in;
			data1_in <= v_data2_in;
		end if;

	end process; -- end ConfilictResolver

	swapAddrRead   <= swapReadUndelayed;
	swapAddrWrite  <= swapWriteUndelayed;
	swapOutputRead <= swapReadDelayed(1);
	mirrorOutput <= mirrorOutputDelayed(1);
	DelaySwap : process(clk)
	begin
	  if rising_edge(clk) then
			if(reset = '1') then
				swapReadDelayed <= (others => '0');
				swapWriteDelayed <= (others => '0');
			else
				swapReadDelayed(1) <= swapReadDelayed(0);
				swapWriteDelayed(1) <= swapWriteDelayed(0);
				mirrorOutputDelayed(1) <= mirrorOutputDelayed(0);
				swapReadDelayed(0) <= swapReadUndelayed;
				swapWriteDelayed(0) <= swapWriteUndelayed;
				mirrorOutputDelayed(0) <= mirrorOutputUndelayed;
			end if;
		end if;
	end process; -- end DelaySwap

	ConflictResolver : process(adr_r, adr_w)
		variable v_swapRead : std_logic;
		variable v_swapWrite : std_logic;
		variable v_mirrorOutput : std_logic;

		variable v_adr1_r : std_logic_vector(adr_width-1 downto 0);
		variable v_adr2_r : std_logic_vector(adr_width-1 downto 0);
		variable v_adr1_w : std_logic_vector(adr_width-1 downto 0);
		variable v_adr2_w : std_logic_vector(adr_width-1 downto 0);
	begin
		v_adr1_r := adr_r(adr_width-1 downto 0);
		v_adr2_r := adr_r(2*adr_width-1 downto adr_width);
		v_adr1_w := adr_w(adr_width-1 downto 0);
		v_adr2_w := adr_w(2*adr_width-1 downto adr_width);

		v_swapRead := '0';
		v_swapWrite := '0';
		v_mirrorOutput := '0';
		if(v_adr1_r = v_adr2_r and v_adr1_w = v_adr2_w and v_adr1_r = v_adr1_w) then
			v_mirrorOutput := '1';
		elsif(v_adr1_w = v_adr1_r and v_adr1_w = v_adr2_r) then
			v_swapWrite := '1';
		elsif(v_adr1_w = v_adr2_r) then
			v_swapRead := '1';
		end if;
		swapReadUndelayed <= v_swapRead;
		swapWriteUndelayed <= v_swapWrite;
		mirrorOutputUndelayed <= v_mirrorOutput;
	end process; -- end ConflictResolver


  Bram_Pump : Bram
  generic map (
    width     => width,
    adr_width => adr_width
  )
  port map (
    clk   => clk,
    reset => reset,
    we   	=> pump_we,
    adr_r => pump_adr_r,
    adr_w => pump_adr_w,
    din  	=> pump_din,
    dout 	=> pump_dout
  );

  portClockGenerator : process(clk)
    variable curState : std_logic;
  begin
    if rising_edge(clk) then
      port1Clk <= '0';
      if(reset = '1') then
        curState := '1';
      else
        if(curState = '0') then
          port1Clk <= '1';
          curState := '1';
        else
          curState := '0';
        end if;
      end if;
    end if;
  end process; -- end portClockGenerator

	ReadOutput : process(doutReg, pump_dout, swapOutputRead)
	begin
		if(mirrorOutput = '1') then
			dout <= doutReg & doutReg;
		else
			if(swapOutputRead = '1') then
				dout <= doutReg & pump_dout;
			else
				dout <= pump_dout & doutReg;
			end if;
		end if;
	end process; -- end ReadOutput

  readSide : process(clk)
  begin
    if rising_edge(clk) then
      if(reset = '1') then
        doutReg <= (others => '0');
      else
        if(port1Clk = '1') then
					doutReg <= pump_dout;
        end if;
      end if;
    end if;
  end process; -- end readSide

  writeSide : process(clk)
  begin
    if rising_edge(clk) then
      pump_we <= '0';
      pump_din <= (others => '0');
      pump_adr_r <= (others => '0');
      pump_adr_w <= (others => '0');
      if(reset = '1') then
      else
        if(port1Clk = '1') then
					pump_adr_r(adr_width-1 DOWNTO 0) <= adr1_r;
					pump_adr_w(adr_width-1 DOWNTO 0) <= adr1_w;
					pump_din(width-1 DOWNTO 0) <= data1_in;

					if(we_inner(0) = '1') then
            pump_we <= '1';
          end if;
        else
					pump_adr_r(adr_width-1 DOWNTO 0) <= adr2_r;
					pump_adr_w(adr_width-1 DOWNTO 0) <= adr2_w;
					pump_din(width-1 DOWNTO 0) <= data2_in;

          if(we_inner(1) = '1') then
            pump_we <= '1';
          end if;
        end if;
      end if;
    end if;
  end process; -- end writeSide
end Bram2x_body;
