library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.OptoUtils.all;

entity MultiPortMem is
    generic (
      m_ports : natural := 8;
      m_width : natural := 48;
      m_adr_width : natural := 8;
      m_pumping : natural := 1
	 );
    Port (
      m_clk : IN STD_LOGIC;
      m_clk2 : IN STD_LOGIC;
			m_reset : IN STD_LOGIC;
			m_enable : IN STD_LOGIC;
			m_we : IN STD_LOGIC_VECTOR(m_ports-1 DOWNTO 0);
			m_read_valid : OUT STD_LOGIC_VECTOR(m_ports-1 DOWNTO 0);
      m_adr_r : IN STD_LOGIC_VECTOR(m_ports*m_adr_width-1 DOWNTO 0);
      m_adr_w : IN STD_LOGIC_VECTOR(m_ports*m_adr_width-1 DOWNTO 0);
      m_data_w : IN STD_LOGIC_VECTOR(m_ports*m_width-1 DOWNTO 0);
      m_data_o : OUT STD_LOGIC_VECTOR(m_ports*m_width-1 DOWNTO 0)
		);
end MultiPortMem;

architecture MultiPortMem_Body of MultiPortMem is
  component Bram2x
  generic (
    width     : natural := 64;
    adr_width : natural := 10
  );
  port (
    clk   : IN  STD_LOGIC;
    reset : IN  STD_LOGIC;
    we    : IN  STD_LOGIC_VECTOR(1 downto 0);
    adr_r : IN  STD_LOGIC_VECTOR(2*adr_width-1 DOWNTO 0);
    adr_w : IN  STD_LOGIC_VECTOR(2*adr_width-1 DOWNTO 0);
    din   : IN  STD_LOGIC_VECTOR(2*width-1 DOWNTO 0);
    dout  : OUT STD_LOGIC_VECTOR(2*width-1 DOWNTO 0)
  );
  end component Bram2x;

  component Bram
  generic (
    width     : natural := 64;
    adr_width : natural := 10
  );
  port (
    clk   : IN  STD_LOGIC;
    reset : IN  STD_LOGIC;
    we    : IN  STD_LOGIC;
    adr_r : IN  STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
    adr_w : IN  STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
    din   : IN  STD_LOGIC_VECTOR(width-1 DOWNTO 0);
    dout  : OUT STD_LOGIC_VECTOR(width-1 DOWNTO 0)
  );
  end component Bram;


  component LVT
  generic (
    ports     : natural := 8;
    adr_width : natural := 10
  );
  port (
    clk    : IN  STD_LOGIC;
    reset  : IN  STD_LOGIC;
    enable : IN  STD_LOGIC;
    we     : IN  STD_LOGIC_VECTOR(ports-1 DOWNTO 0);
    adr_w  : IN  STD_LOGIC_VECTOR(ports * adr_width-1 DOWNTO 0);
    adr_r  : IN  STD_LOGIC_VECTOR(ports * adr_width-1 DOWNTO 0);
    r_sel  : OUT STD_LOGIC_VECTOR(ports * log2(ports+1)-1 DOWNTO 0)
  );
  end component LVT;

    constant innerPortDiv : natural := m_pumping;


    constant rSelWidth : natural := log2(m_ports+1);
    constant singleBusWidth : natural := m_ports*m_width;
    constant mainBusWidth : natural := m_ports*singleBusWidth;
    --subtype data_bus is STD_LOGIC_VECTOR(m_ports*(m_adr_width + m_width)-1 DOWNTO 0);
    subtype data_bus is STD_LOGIC_VECTOR(singleBusWidth-1 DOWNTO 0);
    subtype main_bus is STD_LOGIC_VECTOR(mainBusWidth-1 DOWNTO 0);
    signal innerMainBus : main_bus;
    signal innerSelBus : STD_LOGIC_VECTOR(m_ports * rSelWidth-1 DOWNTO 0);
    signal innerSelBusDelayed  : STD_LOGIC_VECTOR(m_ports * rSelWidth-1 DOWNTO 0);
begin
  assert m_pumping <= 1 report "Pumping can't be larger then 2" severity error;
--  GEN_PORT_BRAMS: for Z in 0 to m_ports-1 generate
--    GEN_BRAMS: for I in 0 to m_ports-1 generate
--      --signal bramx_we : STD_LOGIC_VECTOR((m_width+m_adr_width)/8 -1 DOWNTO 0);
--      signal bramx_we : STD_LOGIC_VECTOR((m_width)/8 -1 DOWNTO 0);
--      --signal bramx_data : STD_LOGIC_VECTOR((m_adr_width + m_width) -1 DOWNTO 0);
--      signal bramx_data : STD_LOGIC_VECTOR((m_width) -1 DOWNTO 0);
--    begin
--        bramx_we <= (others => m_we(I));
--            --bramx_data <= m_adr_w((I+1)*m_adr_width -1  DOWNTO I*m_adr_width) & (m_data_w((I+1)*m_width -1 DOWNTO I*m_width));
--        bramx_data <= m_data_w((I+1)*m_width -1 DOWNTO I*m_width);
--        BRAMX: component Bram
--        generic map(
--            m_width,
--            m_adr_width
--        )
--        port map(
--            clk => m_clk,m_width
--            reset => m_reset,
--            we => bramx_we,
--            adri => m_adr_w((I+1)*m_adr_width -1 DOWNTO I*m_adr_width),
--            din => bramx_data,
--            adro => m_adr_r((Z+1)*m_adr_width -1 DOWNTO Z*m_adr_width),
--            --dout => innerMainBus(Z)((I+1)*(m_adr_width + m_width) -1 DOWNTO I*(m_adr_width + m_width))
--            dout => innerMainBus(Z)((I+1)*(m_width) -1 DOWNTO I*(m_width))
--        );
--     end generate GEN_BRAMS;
--   end generate GEN_PORT_BRAMS;
GEN_PORT_BRAMS: for Z in 0 to m_ports-1 generate
  constant zBusWidth : natural := mainBusWidth/m_ports;
  signal curZBus : std_logic_vector(zBusWidth-1 downto 0);
  begin
    innerMainBus((Z+1)*zBusWidth-1 downto (Z*zBusWidth)) <= curZBus;
    GEN_BRAMS: for I in 0 to (m_ports/innerPortDiv)-1 generate
      constant iBusWidth : natural := zBusWidth/(m_ports/innerPortDiv);
      signal curIBus : std_logic_vector(iBusWidth-1 downto 0);
      signal curWe :  std_logic_vector(innerPortDiv-1 downto 0);
      signal curAdr : std_logic_vector(m_adr_width*innerPortDiv-1 downto 0);
      signal curData : std_logic_vector(m_width*innerPortDiv-1 downto 0);
    begin
      curZBus((I+1)*iBusWidth-1 DOWNTO I*iBusWidth) <= curIBus;
      curWe <= (others => m_we(Z) and m_enable);
      EXTEND: for F in 0 to innerPortDiv-1 generate
        curAdr((F+1)*m_adr_width-1 DOWNTO F*m_adr_width) <= m_adr_w((Z+1)*m_adr_width -1 DOWNTO Z*m_adr_width);
        curData((F+1)*m_width-1 DOWNTO F*m_width) <= m_data_w((Z+1)*m_width -1 DOWNTO Z*m_width);
      end generate;

      --m_we((I+1)*innerPortDiv -1 DOWNTO I*innerPortDiv)
      --m_adr((I+1)*m_adr_width*innerPortDiv -1 DOWNTO I*m_adr_width*innerPortDiv)
      --m_data_w((I+1)*m_width*innerPortDiv -1 DOWNTO I*m_width*innerPortDiv)+

      SimpleBram : if innerPortDiv = 1 generate
          Bram_i : Bram
          generic map (
            width     => m_width,
            adr_width => m_adr_width
          )
          port map (
            clk   => m_clk,
            reset => m_reset,
            we    => curWe(0),
            adr_w => curAdr,
            adr_r => m_adr_r((I+1)*m_adr_width*innerPortDiv -1 DOWNTO I*m_adr_width*innerPortDiv),
            din   => curData,
            dout  => curIBus
          );
      end generate;

      PumpedBram : if innerPortDiv = 2 generate
        Bram2x_i : Bram2x
        generic map (
          width     => m_width,
          adr_width => m_adr_width
        )
        port map (
          clk   => m_clk2,
          reset => m_reset,
          we    => curWe,
          adr_w => curAdr,
          adr_r => m_adr_r((I+1)*m_adr_width*innerPortDiv -1 DOWNTO I*m_adr_width*innerPortDiv),
          din   => curData,
          dout  => curIBus
        );
      end generate;
    end generate GEN_BRAMS;
 end generate GEN_PORT_BRAMS;

     LVTX: component LVT
     generic map(
        m_ports,
        m_adr_width
     )
     port map(
        clk => m_clk,
        reset => m_reset,
        enable => m_enable,
        we => m_we,
        adr_w => m_adr_w,
        adr_r => m_adr_r,
        r_sel => innerSelBus
     );

--     PIPELINEOUTPUT : process(m_clk)
--     begin
--       if rising_edge(m_clk) then
--          if m_reset = '1' then
--            innerSelBusDelayed <= (others => '0');
--          else
--            innerSelBusDelayed <= innerSelBus;
--          end if;
--       end if;
--     end process; -- end PIPELINEOUTPUT

     OUTPUT_MULTPLEXORS : for I in 0 to m_ports-1 generate
      signal curSel  : unsigned(rSelWidth-1 DOWNTO 0);
      signal curBus  : data_bus;
      signal dstData : STD_LOGIC_VECTOR(m_width -1 DOWNTO 0);
      signal lwPortInd : natural := 0;
     begin
       curSel <= unsigned(innerSelBus((I+1)*rSelWidth-1 DOWNTO I*rSelWidth));
       m_read_valid(I) <= '1' when (to_integer(curSel) /= 0) else '0';
       lwPortInd <= to_integer(curSel-1);
       curBus <= innerMainBus((lwPortInd+1)*singleBusWidth-1 DOWNTO lwPortInd*singleBusWidth) when lwPortInd < m_ports else (others => '0');

       dstData <= curBus((I+1)*m_width -1 DOWNTO I*m_width);
       m_data_o((I+1)*m_width -1 DOWNTO I*m_width) <= dstData;
     end generate; -- end for generate OUTPUT_MULTPLEXORS

--     process(innerSelBusDelayed, innerMainBus)
--         variable curSel  : STD_LOGIC_VECTOR(rSelWidth-1 DOWNTO 0);
--         variable curData : STD_LOGIC_VECTOR(m_width -1 DOWNTO 0);
--         variable curBus  : data_bus;
--         variable dstData : STD_LOGIC_VECTOR(m_width -1 DOWNTO 0);
--
--         variable lwPortInd : natural := 0;
--     begin
--       for I in 0 to m_ports-1 loop
--         curSel := innerSelBusDelayed((I+1)*rSelWidth-1 DOWNTO I*rSelWidth);
--         lwPortInd := conv_integer(curSel);
--         curBus := innerMainBus((lwPortInd+1)*singleBusWidth-1 DOWNTO lwPortInd*singleBusWidth);
--
--         dstData := (others => '0');
--         dstData := curBus((I+1)*m_width -1 DOWNTO I*m_width);
--         m_data_o((I+1)*m_width -1 DOWNTO I*m_width) <= dstData;
--       end loop;
--     end process;
end MultiPortMem_Body;
